import click
import sys
from pynif import NIFCollection
from nifconverter.uriconverter import URIConverter
from nifconverter.dbpedia_sparql import ToDBpediaSparqlConverter
from nifconverter.dbpedia_sparql import FromDBpediaSparqlConverter
from nifconverter.dbpedia_samething import SameThingConverter
from nifconverter.translator import NIFTranslator
from nifconverter.utils import WIKIDATA_PREFIX

registered_converters = {cls.__name__: cls for cls in URIConverter.__subclasses__()}

def get_available_converters():
    return ', '.join(registered_converters.keys()) or '(none)'

@click.command()
@click.option('--converter', default='SameThingConverter', help='Conversion method, among the implemented ones: '+get_available_converters())
@click.option('--target', default=WIKIDATA_PREFIX, help='The target namespace to convert the URIs to.')
@click.option('-i', '--infile', default='-', help='The source NIF file to read. If not provided, stdin is used.')
@click.option('-o', '--outfile', default='-', help='The target NIF file to write. If not provided, stdout is used.')
@click.option('--format', default='turtle', help='The RDF serialization format to use.')
def main(converter, target, infile, outfile, format):
    """
    Conversion utility for NIF files.

    This converts the identifiers used to annotate mentions in documents
    across knowledge bases. For instance, the following will convert
    a NIF file with DBpedia identifiers to a NIF file with Wikidata identifiers,
    using the default converter (which uses the DBpedia SameThing service):

       nifconverter -i dbpedia_nif.ttl -o wikidata_nif.ttl

    """

    converter_impl = registered_converters.get(converter)
    if converter_impl is None:
        raise click.BadParameter('Invalid converter "{}". Supported converters are: {}'.format(converter, get_available_converters()))

    translator = NIFTranslator(converter_impl(target_prefix=target))

    with click.open_file(infile) as f:
        nif = NIFCollection.loads(f.read())

    translator.translate_collection(nif)

    with click.open_file(outfile, 'w') as out:
        out.write(nif.dumps())

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from nifconverter.dbpedia_sparql import FromDBpediaSparqlConverter
from nifconverter.dbpedia_sparql import ToDBpediaSparqlConverter
from nifconverter.tests.dbpedia_bases import DBpediaTestBase


class FromDBpediaSparqlConverterTest(DBpediaTestBase.FromDBpediaBase):
    @classmethod
    def setUpClass(cls):
        cls.converter = FromDBpediaSparqlConverter('http://www.wikidata.org/entity/')


class ToDBpediaSparqlConverterTest(DBpediaTestBase.ToDBpediaBase):

    @classmethod
    def setUpClass(cls):
        cls.converter = ToDBpediaSparqlConverter('http://dbpedia.org/resource/')



import itertools

from nifconverter.settings import SAME_THING_SERVICE_URL
from nifconverter.uriconverter import URIConverter
from nifconverter.utils import retry_request, fetch_redirecting_uris

try:
    from urllib.parse import unquote
except ImportError:
    from urllib import unquote


class SameThingConverter(URIConverter):
    """
    URI converter that uses the DBpedia Same Thing Service.
    see http://dev.dbpedia.org/Global_IRI_Resolution_Service
    """
    batch_size = 50

    def __init__(self, target_prefix='http://www.wikidata.org/entity/'):
        super(SameThingConverter, self).__init__(target_prefix)

    def is_convertible(self, uri):
        # todo: https://github.com/dbpedia/dbp-same-thing-service/issues/9
        return any(
            name in uri
            for name in [
                'dbpedia.org',
                'wikidata.org'
            ]
        )

    def convert_one(self, uri):
        """
        Convert one URI (assumed to be convertible).
        Returns the converted URI or None if the concept does
        not exist in the target domain.
        """
        resp = retry_request(
            SAME_THING_SERVICE_URL,
            {
                'meta': 'off',
                'uri': uri
            }
        )
        if resp.status_code == 404:
            # Additionally, try to resolve redirected resources
            redirecting_uris = fetch_redirecting_uris([uri], [None])
            if redirecting_uris:
                return self.convert_one(redirecting_uris[uri])
            else:
                return None

        resp.raise_for_status()
        identifiers = resp.json()
        for target_uri in itertools.chain(
            [identifiers['global']],
            identifiers['locals']
        ):
            if target_uri.startswith(self.target_prefix):
                return target_uri

    def convert(self, uris):
        """
        Converts a list of URIs, whose length should
        not be greater than batch_size.

        Returns a map of results (from the original uri to the target uri).
        """
        resp = retry_request(
            SAME_THING_SERVICE_URL,
            {
                'meta': 'off',
                'uris': uris
            }
        )
        resp.raise_for_status()
        uri_map = resp.json()['uris']
        results = {}

        # try to resolve redirected resources
        redirecting_uris = {}
        if not all(uri_map.values()):
            possible_redirects = [
                source_uri
                for source_uri, identifiers in uri_map.items()
                if not identifiers
            ]
            redirecting_uris = fetch_redirecting_uris(possible_redirects, [None])

        # filter target URIs
        for source_uri, identifiers in uri_map.items():
            if identifiers:
                for target_uri in itertools.chain(
                        [identifiers['global']],
                        identifiers['locals']
                ):
                    if target_uri.startswith(self.target_prefix):
                        results[source_uri] = target_uri
                        break
            elif source_uri in redirecting_uris:
                results[source_uri] = self.convert_one(redirecting_uris[source_uri])

        return results



from nifconverter.uri_converter import URIConverter
from nifconverter.dbpedia_sparql import FromDBpediaSparqlConverter
from nifconverter.dbpedia_sparql import ToDBpediaSparqlConverter
from nifconverter.dbpedia_samething import SameThingConverter

class AutoConverter(URIConverter):
    """
    Meta-converter which selects the best converter given the
    desired source and target prefixes.
    """







class URIConverter(object):
    """
    Interface for a URI converter.
    """
    batch_size = 1

    def __init__(self, target_prefix):
        """
        Initializes the converter to output URIs with the
        given target prefix.
        Subclasses can throw ValueError if the supplied
        target prefix is not supported.
        """
        self.target_prefix = target_prefix

    def is_convertible(self, uri):
        """
        Is this URI convertible by this converter?
        """
        raise NotImplemented

    def convert_one(self, uri):
        """
        Convert one URI (assumed to be convertible).
        Returns the converted URI or None if the concept does
        not exist in the target domain.
        """
        return self.convert([uri]).get(uri)


    def convert(self, uris):
        """
        Converts a list of URIs, whose length should
        not be greater than batch_size.

        Returns a map of results (from the original uri to the target uri).
        """
        raise NotImplemented




# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest

import pytest
import requests_mock
from requests import HTTPError

from nifconverter.dbpedia_samething import SameThingConverter
from nifconverter.settings import SAME_THING_SERVICE_URL
from nifconverter.tests.dbpedia_bases import DBpediaTestBase


class FromDBpSameThingConverterTest(DBpediaTestBase.FromDBpediaBase):
    @classmethod
    def setUpClass(cls):
        cls.converter = SameThingConverter('http://www.wikidata.org/entity/')

    def test_single_redirect(self):
        target_uri = self.converter.convert_one('http://dbpedia.org/page/Sebastiaan_Rutten')
        assert 'http://www.wikidata.org/entity/Q704832' == target_uri


class ToDBpSameThingConverterTest(DBpediaTestBase.ToDBpediaBase):
    unconvertible_uris = [
        'mock://not.an.URI',
        'mock://bad.domain/surprise',
        'mock://www.wikidata.org/entity/null',
    ]

    @classmethod
    def setUpClass(cls):
        cls.converter = SameThingConverter('http://dbpedia.org/resource/')

    def test_some_uris_unconvertible(self):
        some_bad_uris = self.unconvertible_uris + ['http://www.wikidata.org/entity/Q1985']
        expected_mapping = {'http://www.wikidata.org/entity/Q1985': 'http://dbpedia.org/resource/2000'}
        with requests_mock.Mocker() as req_mocker:
            req_mocker.get('mock://', status_code=404)
            req_mocker.get(SAME_THING_SERVICE_URL, real_http=True)
            mapping = self.converter.convert(some_bad_uris)

        self.assertEqual(expected_mapping, mapping)

    def test_all_uris_unconvertible(self):
        with pytest.raises(HTTPError) as excinfo:
            self.converter.convert(self.unconvertible_uris)

        assert '404' in str(excinfo.value)


class MultilingualDBpSameThingConverterTest(unittest.TestCase):
    source_uri = 'http://www.wikidata.org/entity/Q8087'

    def test_global(self):
        stc = SameThingConverter('https://global.dbpedia.org/id/')
        target_uri = stc.convert_one(self.source_uri)
        self.assertEqual('https://global.dbpedia.org/id/4y9Et', target_uri)

    def test_eo(self):
        stc = SameThingConverter('http://eo.dbpedia.org/resource/')
        target_uri = stc.convert_one(self.source_uri)
        self.assertEqual('http://eo.dbpedia.org/resource/Geometrio', target_uri)

    def test_nn(self):
        stc = SameThingConverter('http://nn.dbpedia.org/resource/')
        target_uri = stc.convert_one(self.source_uri)
        self.assertEqual('http://nn.dbpedia.org/resource/Geometri', target_uri)

    def test_mk(self):
        stc = SameThingConverter('http://mk.dbpedia.org/resource/')
        target_uri = stc.convert_one(self.source_uri)
        self.assertEqual('http://mk.dbpedia.org/resource/Геометрија', target_uri)

    def test_af(self):
        stc = SameThingConverter('http://af.dbpedia.org/resource/')
        target_uri = stc.convert_one(self.source_uri)
        self.assertEqual('http://af.dbpedia.org/resource/Meetkunde', target_uri)

    def test_is(self):
        stc = SameThingConverter('http://is.dbpedia.org/resource/')
        target_uri = stc.convert_one(self.source_uri)
        self.assertEqual('http://is.dbpedia.org/resource/Rúmfræði', target_uri)




import requests_mock
import pytest
import requests
from nifconverter.utils import retry_request, get_redirect


def test_retry_request_error():
    with requests_mock.mock() as mocker:
        mocker.get('http://somehost.sometld/', status_code=500)
        with pytest.raises(requests.RequestException):
            retry_request('http://somehost.sometld/', delay=0)


def test_retry_request_successful():
    with requests_mock.mock() as mocker:
        mocker.get('http://somehost.sometld/', text='hello')
        req = retry_request('http://somehost.sometld/', delay=0)
        assert req.text == 'hello'


def test_redirects_to():
    assert (
        get_redirect('http://dbpedia.org/page/Gio_Gonzalez')
        == 'http://dbpedia.org/page/Gio_Gonz%C3%A1lez'
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
try:
    from urllib.parse import unquote
except ImportError:
    from urllib import unquote

from .uriconverter import URIConverter
from .utils import retry_request, fetch_redirecting_uris, DBPEDIA_PREFIX, DBPEDIA_PAGE_PREFIX


class FromDBpediaSparqlConverter(URIConverter):
    batch_size = 20

    def __init__(self, target_prefix='http://en.wikipedia.org/wiki/'):
        """
        Creates a converter from DBpedia to one of the other URI schemes
        DBpedia knows of via owl:sameAs.

        The target prefix is used to select these links.
        """
        super(FromDBpediaSparqlConverter, self).__init__(target_prefix)

    def is_convertible(self, uri):
        return uri.startswith(DBPEDIA_PREFIX) or uri.startswith(DBPEDIA_PAGE_PREFIX)

    def convert(self, uris):
        """
        This uses DBpedia's SPARQL endpoint to convert the identifiers.
        """
        decoded_uris = {
            uri: unquote(uri).replace(' ', '_').replace('"', '%22').replace(DBPEDIA_PAGE_PREFIX, DBPEDIA_PREFIX)
            for uri in uris
        }

        sparql_query = """
        SELECT ?source ?target WHERE {{
           ?source owl:sameAs ?target .
           VALUES ?source {{ {uris} }}
        }}
        """.format(uris=' '.join({'<{}>'.format(uri) for uri in decoded_uris.values()}))

        mapping = fetch_mapping(sparql_query, self.target_prefix)

        # Additionally, try to resolve redirected resources
        redirecting_uris = fetch_redirecting_uris(decoded_uris.values(), mapping.keys())
        if redirecting_uris:
            redirect_mapping = self.convert(redirecting_uris.values())
            mapping.update({
                decoded_uri: redirect_mapping[redirected_uri]
                for decoded_uri, redirected_uri in redirecting_uris.items()
                if redirected_uri in redirect_mapping
            })

        return {
            uri: mapping[decoded_uri]
            for uri, decoded_uri in decoded_uris.items()
            if decoded_uri in mapping
        }


class ToDBpediaSparqlConverter(URIConverter):
    batch_size = 20

    def __init__(self, target_prefix=DBPEDIA_PREFIX):
        """
        Creates a converter to DBpedia to one of the other URI schemes
        DBpedia knows of via owl:sameAs.
        """
        super(ToDBpediaSparqlConverter, self).__init__(DBPEDIA_PREFIX)
        if target_prefix != DBPEDIA_PREFIX:
            raise ValueError('ToDBpediaSparqlConverter only supports converting to DBpedia URIs, starting with '+DBPEDIA_PREFIX)

    def is_convertible(self, uri):
        # TODO: find out which namespaces are available from the SPARQL interface
        return True

    def convert(self, uris):
        """
        This uses DBpedia's SPARQL endpoint to convert the identifiers.
        """
        uris = [uri.replace(' ', '_').replace('"', '%22') for uri in uris]

        sparql_query = """
        SELECT ?source ?target WHERE {{
           ?target owl:sameAs ?source .
           VALUES ?source {{ {uris} }}
        }}
        """.format(uris=' '.join('<{}>'.format(uri) for uri in uris))

        return fetch_mapping(sparql_query, DBPEDIA_PREFIX)


def fetch_mapping(sparql_query, prefix):
    r = retry_request(
        'http://dbpedia.org/sparql/',
        {
            'query': sparql_query,
            'format': 'json'
        }
    )
    r.raise_for_status()
    results = r.json()['results']

    mapping = {}
    for binding in results['bindings']:
        source = binding['source']['value']
        target = binding['target']['value']
        if target.startswith(prefix):
            mapping[source] = target

    return mapping



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import requests
from time import sleep

logger = logging.getLogger(__name__)

DBPEDIA_PREFIX = 'http://dbpedia.org/resource/'
DBPEDIA_PAGE_PREFIX = 'http://dbpedia.org/page/'
WIKIDATA_PREFIX = 'http://www.wikidata.org/entity/'

def retry_request(url, parameters=None, timeout=10, delay=10, retries=4):
    """
    Requests an URL with retries
    :returns: the Request object if succeeded
    """
    parameters = parameters or {}
    backoff = 2
    while retries:
        try:
            req = requests.get(url, parameters, timeout=timeout)
            if req.status_code != 404:
                req.raise_for_status()
            return req
        except requests.RequestException as e:
            logger.info('Retrying in {} seconds for {}'.format(delay, url))
            sleep(delay)
            retries -= 1
            delay *= backoff
            if not retries:
                raise


def get_redirect(url):
    """
    Checks if a URL redirects to another URL, in
    which case the new URL is returned. Otherwise None is returned.
    """
    # Sadly a HEAD request does not work for obscure encoding reasons...
    # See accompanying test case
    if not url.startswith('http'):
        return
    try:
        req = retry_request(url, delay=3, retries=1)
    except requests.exceptions.RequestException:
        return
    location = req.url
    if location and location != url:
        return location


def fetch_redirecting_uris(decoded_uris, mapped_uris):
    missing_dbps = set(decoded_uris) - set(mapped_uris)
    redirecting_uris = {}
    for missing_uri in missing_dbps:
        redirect = get_redirect(missing_uri.replace(DBPEDIA_PREFIX, DBPEDIA_PAGE_PREFIX))
        if redirect:
            redirecting_uris[missing_uri] = redirect

    return redirecting_uris






import os
from nifconverter.cli import main
import pynif
from click.testing import CliRunner

def test_invoke_cli():
    testdir = os.path.dirname(os.path.abspath(__file__))
    dbpedia_nif_path = os.path.join(testdir, 'data/sample_dbpedia.ttl')
    wikidata_nif_path = os.path.join(testdir, 'data/sample_wikidata.ttl')
    target_file = os.path.join(testdir, 'out.ttl')

    runner = CliRunner()
    result = runner.invoke(main, ['-i', dbpedia_nif_path])
    assert result.exit_code == 0

    nif = pynif.NIFCollection.loads(result.output)
    expected_nif = pynif.NIFCollection.load(wikidata_nif_path)
    assert expected_nif == nif






class NIFTranslator(object):
    """
    Takes NIFCollection object and translates
    it using a URI converter.
    """

    def __init__(self, uri_converter):
        self.uri_converter = uri_converter

    def translate_collection(self, collection):
        """
        Given a collection, add the new URIs using the configured converter.
        Set remove_source_uris to True if the original URIs should be replaced (otherwise they are kept).

        The collection is modified in place.
        """
        all_phrase_uris = [phrase.taIdentRef for phrase in self._enumerate_phrases(collection) if phrase.taIdentRef]
        convertible_uris = {uri for uri in all_phrase_uris if self.uri_converter.is_convertible(uri)}

        batches = self._batchify(convertible_uris, self.uri_converter.batch_size)

        uri_map = {}
        for batch in batches:
            uri_map.update(self.uri_converter.convert(batch))

        for phrase in self._enumerate_phrases(collection):
            if uri_map.get(phrase.taIdentRef):
                phrase.taIdentRef = uri_map[phrase.taIdentRef]

    def _enumerate_phrases(self, collection):
        """
        Enumerates all the phrases in a collection
        """
        for doc in collection.contexts:
            for phrase in doc.phrases:
                yield phrase

    def _batchify(self, enumerable, batch_size):
        """
        Cuts an enumerable into a series of batches of size at most batch_size.
        """
        current_batch = []
        for item in enumerable:
            current_batch.append(item)
            if len(current_batch) == batch_size:
                yield current_batch
                current_batch = []
        if current_batch:
            yield current_batch



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import unittest
from nifconverter.uriconverter import URIConverter

try:
    from urllib.parse import quote
except ImportError:
    from urllib import quote


class DBpediaTestBase:
    # class nesting prevents the test runner from picking up bases

    class FromDBpediaBase(unittest.TestCase):

        @classmethod
        def setUpClass(cls):
            cls.converter = URIConverter()

        def test_is_convertible(self):
            self.assertTrue(self.converter.is_convertible('http://dbpedia.org/resource/Douglas_Adams'))
            self.assertFalse(self.converter.is_convertible('http://en.wikipedia.org/wiki/Douglas_Adams'))
            self.assertTrue(self.converter.is_convertible('http://dbpedia.org/page/Nebraska_Cornhuskers_football'))

        def test_to_wikidata(self):
            expected_mapping = {
                # Simple case
                'http://dbpedia.org/resource/Douglas_Adams': 'http://www.wikidata.org/entity/Q42',
                # With alternate prefix
                'http://dbpedia.org/page/Nebraska_Cornhuskers_football':'http://www.wikidata.org/entity/Q6984693',
                # For escaping
                'http://dbpedia.org/resource/Fran%C3%A7ois_Legault': 'http://www.wikidata.org/entity/Q3085147',
                # Without escaping, with unicode character
                'http://dbpedia.org/resource/François_Legault': 'http://www.wikidata.org/entity/Q3085147',
                # Without escaping, with quotes
                'http://dbpedia.org/page/Toys_%22R%22_Us': 'http://www.wikidata.org/entity/Q696334',
                # With redirect
                'http://dbpedia.org/resource/Gio_Gonzalez': 'http://www.wikidata.org/entity/Q1525217',
            }
            mapping = self.converter.convert(expected_mapping.keys())
            self.assertEqual(expected_mapping, mapping)

    class ToDBpediaBase(unittest.TestCase):

        @classmethod
        def setUpClass(cls):
            cls.converter = URIConverter()

        def test_is_convertible(self):
            self.assertTrue(self.converter.is_convertible('http://www.wikidata.org/entity/Q42'))

        def test_from_wikidata(self):
            mapping = self.converter.convert([
                'http://www.wikidata.org/entity/Q42',
                'http://www.wikidata.org/entity/Q34433'
            ])
            expected_mapping = {
               'http://www.wikidata.org/entity/Q34433': 'http://dbpedia.org/resource/University_of_Oxford',
               'http://www.wikidata.org/entity/Q42': 'http://dbpedia.org/resource/Douglas_Adams',
            }
            self.assertEqual(expected_mapping, mapping)



import os
import unittest
from nifconverter.translator import NIFTranslator
from nifconverter.uriconverter import URIConverter
from pynif import NIFCollection

class DummyURIConverter(URIConverter):
    dbp_prefix = 'http://dbpedia.org/resource/'
    wp_prefix = 'http://en.wikipedia.org/wiki/'

    def __init__(self):
        super(DummyURIConverter, self).__init__(self.wp_prefix)

    def is_convertible(self, uri):
        return uri.startswith(self.dbp_prefix)

    def convert_one(self, uri):
        return self.wp_prefix + uri[len(self.dbp_prefix):]

    def convert(self, uris):
        return {uri:self.convert_one(uri) for uri in uris}


class NIFTranslatorTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        testdir = os.path.dirname(os.path.abspath(__file__))
        cls.dbpedia_nif = NIFCollection.load(os.path.join(testdir, 'data/sample_dbpedia.ttl'))
        cls.wikipedia_nif = NIFCollection.load(os.path.join(testdir, 'data/sample_wikipedia.ttl'))

    def test_translate_collection(self):
        converter = DummyURIConverter()
        translator = NIFTranslator(converter)

        translator.translate_collection(self.dbpedia_nif)

        self.assertEqual(self.wikipedia_nif, self.dbpedia_nif)

